// HSolidView.cpp : implementation of the HSolidView class
// 

#include "HSolidView.h"
#include "HSolidModel.h"
#include "HSSelectionSet.h"

#include "HSUtility.h"
#include <assert.h>
#include "hc.h"

#ifdef PARASOLID
#include "hp_bridge.h"
#endif
#ifdef ACIS
#include "ha_bridge.h"
// Required to use constant  FACE_TYPE
#include "kernel/kerndata/top/face.hxx"
// Required to use constant  BODY_TYPE
#include "kernel/kerndata/top/body.hxx"
// Required to use constant  EDGE_TYPE
#include "kernel/kerndata/top/edge.hxx"

#endif

#define MAX_BOOLEAN_TOOLS 1024


HSolidView::HSolidView(HBaseModel *model,
    char * 		alias,	
    char *		driver_type,
	char *		instance_name,
    long		window_handle,
    long		colormap) : HBaseView(model, alias, driver_type, instance_name, window_handle, colormap)

{
	m_optimize_shells = false;
}

// app-specific init function
void HSolidView::Init()
{
	HSolidModel * model = (HSolidModel *)GetModel();

	// call base's init function first to get the default HOOPS hierarchy for the view
	HBaseView::Init();

	// create our app-specific Selection object and initialize
	m_pSelection = new HSSelectionSet(this);
    m_pSelection->Init();

    // app-specific scene Defaults

    HC_Open_Segment_By_Key(m_ViewKey);
		HC_Set_Color_By_Index("windows", 0);
		HC_Set_Selectability("everything = off");
	HC_Close_Segment();

	// set up some scene defaults
	HC_Open_Segment_By_Key(m_SceneKey);

	  HC_Set_Color_By_Index("faces", 2);
	  HC_Set_Color_By_Index("text, lights", 1);
	  HC_Set_Color_By_Index("edges, lines", 1);
	  HC_Set_Color_By_Index("markers", 1);
	  HC_Set_Rendering_Options("no color interpolation, color index interpolation");
	  HC_Set_Visibility ("lights = (faces = on, edges = off), markers = off, faces=on, edges=off, lines=off, text = on");

	  m_parts_seg = HC_KCreate_Segment("parts");
	  
	HC_Close_Segment();

    // windowspace (overlay) defaults
    HC_Open_Segment_By_Key(m_WindowspaceKey);
	  HC_Set_Color_By_Index ("geometry", 3);
	  HC_Set_Color_By_Index ("window contrast", 1);
	  HC_Set_Color_By_Index ("windows", 1);

	  HC_Set_Visibility("markers=on");
	  HC_Set_Marker_Symbol("+");
	  HC_Set_Selectability("off");
    HC_Close_Segment();	
}


void HSolidView::DeleteSelectionList()
{
	int				i, numSolidSelections, numHoopsSelections;
	char			type[256];
	long			tmp;
	HSSelectionSet* selection = 0;
#ifdef PARASOLID
	PK_ENTITY_t		current = 0;
#endif
#ifdef ACIS
	ENTITY*			current = 0;
#endif

	// get ptr to selection object
	selection = (HSSelectionSet *)GetSelection();

	numHoopsSelections = selection->GetSize();
	numSolidSelections = selection->GetSolidListSize();

	if (numHoopsSelections == 0)
		return;

	// loop through the list of selected HOOPS primitives, filtering
	// out the ones associated with solid model entities, and deleting
	// the non-solid model primitives directly using Delete_By_Key
	// The primitives associated with solid model entities need to be
	// deleted using the HOOPS/GM deletion functions so that the HOOPS<->GM 
	// mapping tables will remain in sync

	for (i = numHoopsSelections; i > 0; i--)
	{
		long key = selection->GetAt(i-1);

		// if it's a HOOPS primitive that is not associated with a solid model entity,
		// deselect and delete it
#ifdef PARASOLID
		if(current = HP_Compute_TagID(key, PK_CLASS_body) <= 0 )
#endif
#ifdef ACIS
		current = HA_Compute_Entity_Pointer(key, BODY_TYPE);
		if (current == 0)
#endif
		{
			selection->DeSelect(key);
			HC_Delete_By_Key(key);
		}
	}


	HC_Open_Segment_By_Key(m_CuttingPlanesKey);
		HC_Begin_Contents_Search ("...", "cutting planes");
		if (!HC_Find_Contents (type, &tmp)) 
			SetCuttingPlanesExist(false); 
		HC_End_Contents_Search ();
	HC_Close_Segment();

    m_bUpdateCPGeom = true;
	

	// deselect the HOOPS primitives
	selection->HSelectionSet::DeSelectAll();

	// loop through the list of selected solid model entities,
	// delete the HOOPS geometry associated with each
	// and then delete the entity
	for (i = numSolidSelections; i > 0; i--)
	{
#ifdef PARASOLID
		current = selection->GetAtSolidEntity(i-1);
		if (current)
		{
			((HSolidModel*)GetModel())->DeleteParasolidEntity(current);
		}
#endif

#ifdef ACIS
		current = (ENTITY*)selection->GetAtSolidEntity(i-1);
		if (current)
		{
			((HSolidModel*)GetModel())->DeleteAcisEntity(current);
			((HSolidModel*)GetModel())->RemoveAcisEntity(current);
		}
#endif

	}

	// Reset the HSSelectionSet and HSelectionSet lists since we've deleted all the entities that
	// were in them  (we can't DeSelect the entities because they no longer exist)
	selection->Reset();

	Update();
}



// perform a boolean operation on selected items
void HSolidView::BooleanOperation(unsigned long	operation)
{
	int			j, count = 0;
	int			number_of_tools = 0;
#ifdef PARASOLID
	PK_BODY_t	target = 0;
	PK_BODY_t	tools[MAX_BOOLEAN_TOOLS];
#endif
#ifdef ACIS
	BODY*		target = 0;
	BODY*		tools[MAX_BOOLEAN_TOOLS];
	ENTITY_LIST elist;
#endif

	int numSelections;
	HSSelectionSet* selection = 0;

	// get ptr to selection object
	selection = (HSSelectionSet *)GetSelection();
	assert(selection);

	// get the number of selected items
	numSelections = selection->GetSolidListSize();
	number_of_tools = numSelections - 1;

	// must have at least 2 selected items for a boolean operation
	if (numSelections < 2)
		return;
#ifdef PARASOLID
	target = (PK_BODY_t)selection->GetAtSolidEntity(0);

	for (j = 0; j < number_of_tools && number_of_tools < MAX_BOOLEAN_TOOLS; j++)
	{
		tools[j] = (PK_BODY_t)selection->GetAtSolidEntity(j+1);
	}

	// deselect all the HOOPS primitives and GM entities before deleting
	selection->DeSelectAll();

	// now blow away ALL of the geometry associated with target
	HP_Delete_Entity_Geometry(1, &target);
	HP_Delete_Entity_Geometry(number_of_tools, tools);
#endif
#ifdef ACIS
	target = (BODY*)selection->GetAtSolidEntity(0);

	for (j = 0; j < number_of_tools && number_of_tools < MAX_BOOLEAN_TOOLS; j++)
	{
		tools[j] = (BODY*)selection->GetAtSolidEntity(j+1);
		((HSolidModel*)GetModel())->RemoveAcisEntity(tools[j]);  // remove tools from list of entities in the model
	}


	// deselect all the HOOPS primitives and GM entities before deleting
	selection->DeSelectAll();

	// now blow away ALL of the geometry associated with target
	elist.add(target);
	for (j = 0; j < number_of_tools; j++)
	{
		elist.add(tools[j]);
	}
	HA_Delete_Entity_Geometry(elist);
#endif

		


    // Perform the boolean operation and render the resulting entities
	// Put each entity in a separate segment so that we can easily apply
	// transformations to separately created entities.
	// However, before rendering bodies, check if this view is in 'merge faces' mode or not, and
	// ensure tha the rendering option is set accordingly.  This is necessary because the
	// HOOPS/GM Rendering Options are procedural; they set the current state of the bridge,
	// and we need to make sure that all parts in each model object are rendered in a 
	// consistent mode
	HC_Open_Segment_By_Key(GetModel()->GetModelKey());
		HC_Open_Segment("");
#ifdef ACIS
			if (m_optimize_shells)
				HA_Set_Rendering_Options("merge faces");
			else
				HA_Set_Rendering_Options("no merge faces");

			acis_boolean(operation, target, number_of_tools, tools);
#endif
#ifdef PARASOLID

			if (m_optimize_shells)
				HP_Set_Rendering_Options("merge faces");
			else
				HP_Set_Rendering_Options("no merge faces");

			parasolid_boolean(operation, target, number_of_tools, tools);
#endif
		HC_Close_Segment();
	HC_Close_Segment();

	m_bUpdateCPGeom = true;

    Update();	
}



void HSolidView::RegenerateBodies(void)
{
#ifdef PARASOLID
	PK_BODY_t body;
#endif
#ifdef ACIS
	ENTITY* entity;
#endif
	HC_KEY	child;
	char seg_type[200];

	HC_Begin_Contents_Search (".", "segment");
	while (HC_Find_Contents (seg_type, &child) )
	{
		if (strcmp(seg_type, "segment"))
		{
			assert(0);
			continue;
		}
#ifdef PARASOLID
		body = HP_Compute_TagID(child, PK_CLASS_body);
		if (body > 0)
		{
			HP_Delete_Entity_Geometry(1, &body);
			HP_Render_Entity (body);
			continue;
		}
#endif
#ifdef ACIS
		entity = HA_Compute_Entity_Pointer(child, BODY_TYPE);
		if (entity)
		{
			ENTITY_LIST elist;

			elist.add(entity);
			HA_Delete_Entity_Geometry(elist);
			HA_Render_Entity (entity);
			continue;
		}
#endif
		HC_Open_Segment_By_Key(child);
			RegenerateBodies();
		HC_Close_Segment();
	}
	HC_End_Contents_Search();
}

void HSolidView::SetOptimizations(bool optimize)
{
	if (optimize != m_optimize_shells)
	{
		m_optimize_shells = optimize;

		// if optimize is true, set 'merge faces' mode so that all faces in a body will
		// be merged into a single HOOPS shell primitive; also set the select level to body
		// level if it was previously face-level, since we can no longer select on 
		// individiual faces

#ifdef PARASOLID
		PK_CLASS_t class_type;

		if (optimize)
		{
			HP_Set_Rendering_Options("merge faces");

			class_type = ((HSSelectionSet *)GetSelection())->GetSelectLevel();
			if (class_type == PK_CLASS_face)
				((HSSelectionSet *)GetSelection())->SetSelectLevel(PK_CLASS_body);
		}
		else
			HP_Set_Rendering_Options("no merge faces");
#endif

#ifdef ACIS
		int class_type;

		if (optimize)
		{
			HA_Set_Rendering_Options("merge faces");

			class_type = ((HSSelectionSet *)GetSelection())->GetSelectLevel();
			if (class_type == FACE_TYPE)
				((HSSelectionSet *)GetSelection())->SetSelectLevel(BODY_TYPE);
		}
		else
			HA_Set_Rendering_Options("no merge faces");
#endif

		// deselect all geometry prior to regenerating the model
		GetSelection()->DeSelectAll();
	
		HC_Open_Segment_By_Key(GetModel()->GetModelKey());
			RegenerateBodies();
		HC_Close_Segment();
	}

}


bool HSolidView::GetOptimizations(void)
{
	return m_optimize_shells;
}
